import 'dart:convert';

import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:firebase_auth/firebase_auth.dart';
import 'package:flutter/material.dart';
import 'package:puma_firebase/entities/condition.dart';
import 'package:puma_firebase/helpers/firestore_helper.dart';

// Puma Firestore.
class PumaFirestore {
  static synchronizeDocuments({
    required String collectionPath,
    required PumaCondition primaryCondition,
    void Function(List<String>)? onCompleted,
    List<PumaCondition>? conditions,
  }) {
    if (FirebaseAuth.instance.currentUser == null) return;
    bool foundDocuments = false;

    CollectionReference<Map<String, dynamic>> collection =
        FirebaseFirestore.instance.collection(collectionPath);

    Query<Map<String, dynamic>> query =
        queryCollection(collection, primaryCondition);

    if (conditions != null) {
      query = queryConditions(query, conditions);
    }

    var snapshot = query.get();

    snapshot.catchError((e) {
      debugPrint("Something went wrong");
      debugPrint(e.toString());
    });

    snapshot.then((value) {
      List<String> documents = [];
      foundDocuments = true;

      for (var doc in value.docs) {
        documents.add(jsonEncode(doc.data()));
      }

      if (onCompleted != null) {
        onCompleted(documents);
      }
    });

    snapshot.whenComplete(() {
      if (!foundDocuments) {
        debugPrint("No documents found in Firestore");
        if (onCompleted != null) {
          onCompleted([]);
        }
      }
    });
  }

  static uploadDocument({
    required String collection,
    required String uuid,
    required String object,
    bool force = false,
  }) {
    if (FirebaseAuth.instance.currentUser == null && !force) return;

    CollectionReference col = FirebaseFirestore.instance.collection(collection);

    col.doc(uuid).set(json.decode(object));
  }

  static getDocument({
    required String collection,
    required String uuid,
    required void Function(String?) onCompleted,
  }) {
    if (FirebaseAuth.instance.currentUser == null) return;

    CollectionReference col = FirebaseFirestore.instance.collection(collection);

    col.doc(uuid).get().then((doc) {
      Object? obj = doc.data();
      String? res;

      if (obj != null) {
        res = jsonEncode(doc.data());
      }

      onCompleted(res);
    });
  }

  static deleteDocument({
    required String collection,
    required String uuid,
  }) {
    if (FirebaseAuth.instance.currentUser == null) return;

    CollectionReference col = FirebaseFirestore.instance.collection(collection);

    col.doc(uuid).delete();
  }

  static listenToDocumentChanges({
    required String collectionPath,
    required PumaCondition primaryCondition,
    void Function(List<String>)? onChanged,
    void Function(String)? onAdded,
    void Function(String)? onModified,
    void Function(String)? onRemoved,
    List<PumaCondition>? conditions,
  }) {
    if (FirebaseAuth.instance.currentUser == null) return;

    CollectionReference<Map<String, dynamic>> collection =
        FirebaseFirestore.instance.collection(collectionPath);

    Query<Map<String, dynamic>> query =
        queryCollection(collection, primaryCondition);

    if (conditions != null) {
      query = queryConditions(query, conditions);
    }

    Stream stream = query.snapshots();

    stream.listen(
      (querySnapshot) {
        List<String> documents = [];

        querySnapshot.docs.forEach((doc) {
          documents.add(jsonEncode(doc.data()));
        });

        if (onChanged != null) {
          onChanged(documents);
        }

        querySnapshot.docChanges.forEach(
          (res) {
            if (res.type == DocumentChangeType.added) {
              debugPrint("Added");
              if (onAdded != null) onAdded(jsonEncode(res.doc.data()));
            } else if (res.type == DocumentChangeType.modified) {
              debugPrint("modified");
              if (onModified != null) onModified(jsonEncode(res.doc.data()));
            } else if (res.type == DocumentChangeType.removed) {
              debugPrint("removed");
              if (onRemoved != null) onRemoved(jsonEncode(res.doc.data()));
            }
          },
        );
      },
    );
  }
}
